#version 330

#define BRIGHTNESS_FACTOR 1.0
#define CELSHADING true
#define CELSHADING_SHADES 15

in vec4 positionFrag;
in vec4 normalFrag;
in vec2 uvFrag;

// coordinates in the ShadowMap
in vec4 SM_tex_coord;

// material information
uniform vec3 ambientFactor;
uniform vec3 diffuseFactor;
uniform vec3 specularFactor;
uniform float shininess;

// point-lights (lanterns)
#define MAX_LANTERNLIGHTS 10
uniform vec3 pointLightPositions[MAX_LANTERNLIGHTS];
uniform vec3 pointLightColors[MAX_LANTERNLIGHTS];
uniform float pointLightIntensities[MAX_LANTERNLIGHTS];
uniform float pointLightRanges[MAX_LANTERNLIGHTS];

// directional light (sun)
uniform vec3 lightVector;

// the camera's position
uniform vec4 cameraPosition;

uniform sampler2D colorTexture;

// shadow map
uniform sampler2D shadow_map;

// the asteroids shadow
uniform sampler2D asteroid_shadow;

out vec4 fragColor;

void main() {
	
	vec4 lightDirection = normalize(vec4(lightVector, 0.0));
	vec4 viewVector = normalize(positionFrag - cameraPosition);



	///// VARIANCE SHADOW-MAPPING FOR SUN-LIGHT /////
	
	float visibility = 1.0;
	
	if(SM_tex_coord.w > 0.0) {
		
		vec3 SM_position = SM_tex_coord.xyz / SM_tex_coord.w;
		
		// check if SM_position is in range
		if (SM_position.x >= 0.0 && SM_position.y >= 0.0 && SM_position.z >= 0.0 && SM_position.x < 1.0 && SM_position.y < 1.0 && SM_position.z < 1.0) {

			vec2 moments = texture(shadow_map, SM_position.xy).rg;
			
			if (SM_position.z > moments.x) {

				float variance = max(moments.y - (moments.x * moments.x), 0.00002);
				
				float diff = SM_position.z - moments.x;

				visibility = variance / (variance + (diff * diff));
				visibility = smoothstep(0.0, 1.0, visibility);

			}

		}

	}
	


	///// SUN-LIGHT /////

	vec3 color = texture(colorTexture, uvFrag).rgb;
	
	vec4 reflectedLightVector = reflect(lightDirection, normalFrag);

	float k_diffuse = max(dot(normalFrag, lightDirection), 0.0);
	float k_specular = max(dot(viewVector, reflectedLightVector), 0.0);

	vec3 diffuseColor = k_diffuse * diffuseFactor;
	vec3 specularColor = pow(k_specular, shininess) * specularFactor * k_diffuse;
	
	vec3 intensity = (ambientFactor + (diffuseColor + specularColor) * visibility);

	fragColor = vec4(color.x * intensity.x, color.y * intensity.y, color.z * intensity.z, 1.0f);

	

	///// LANTERN LIGHTS /////

	for (int i = 0; i < MAX_LANTERNLIGHTS; i++) {

		if (pointLightIntensities[i] > 0.0f && dot(normalize(pointLightPositions[i]), normalize(lightVector)) < -0.25f) {
			
			float lightLength = length(vec3(positionFrag) - pointLightPositions[i]);

			if (lightLength < pointLightRanges[i]) {
				
				vec4 lightDirection_PointLight = normalize(vec4(pointLightPositions[i], 1.0) - positionFrag);
				vec4 reflectedLightVector_PointLight = reflect(lightDirection_PointLight, normalFrag);
					
				float k_diffuse_PointLight = max(dot(normalFrag, lightDirection_PointLight), 0.0);
				float k_specular_PointLight = max(dot(viewVector, reflectedLightVector_PointLight), 0.0) * k_diffuse_PointLight;
				
				vec3 diffuseColor_PointLight = k_diffuse_PointLight * diffuseFactor;
				vec3 specularColor_PointLight = pow(k_specular_PointLight, shininess) * specularFactor;
				
				vec3 intensity_PointLight = (diffuseColor_PointLight + specularColor_PointLight) * pointLightIntensities[i] * (1.0f - lightLength/pointLightRanges[i]);

				vec3 color_PointLight = vec3(color.x * pointLightColors[i].x, color.y * pointLightColors[i].y, color.z * pointLightColors[i].z) * intensity_PointLight;

				fragColor = vec4(fragColor.x + color_PointLight.x, fragColor.y + color_PointLight.y, fragColor.z + color_PointLight.z, 1.0f);
				
			}
			
		}

	}
	

	
	///// MARK ASTEROIDS THAT ARE NEAR THE PLANET IN RED /////
	
	float distanceToPlanet = length(positionFrag) - 1.0;

	float redFacor = (distanceToPlanet - 2.1) * -0.8;
	redFacor = pow(redFacor, 4.0);

	fragColor.x = fragColor.x + (fragColor.x * redFacor);
	


	///// CEL-SHADING /////
	
	if (CELSHADING) {
		
		fragColor.x = float(int((fragColor.x) * CELSHADING_SHADES)) / CELSHADING_SHADES;
		fragColor.y = float(int((fragColor.y) * CELSHADING_SHADES)) / CELSHADING_SHADES;
		fragColor.z = float(int((fragColor.z) * CELSHADING_SHADES)) / CELSHADING_SHADES;

	}
	
	fragColor.w = 1.0;
	
	fragColor = fragColor * BRIGHTNESS_FACTOR;

}